@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # flake8: noqa
-"""Top-level module for dask_groupby ."""
-from .core import groupby_reduce  # noqa
+"""Top-level module for flox ."""
+from .core import groupby_reduce, rechunk_for_blockwise, rechunk_for_cohorts  # noqa
 
 try:
     from importlib.metadata import version as _version
@@ -10,7 +10,7 @@ except ImportError:
     from importlib_metadata import version as _version  # type: ignore[no-redef]
 
 try:
-    __version__ = _version("dask_groupby")
+    __version__ = _version("flox")
 except Exception:
     # Local copy or not installed with setuptools.
     # Disable minimum version checks on downstream libraries.
