@@ -1,7 +1,6 @@
 import itertools
 from typing import TYPE_CHECKING, Hashable, Iterable, Optional, Sequence, Tuple, Union
 
-import dask
 import numpy as np
 import pandas as pd
 import xarray as xr
@@ -10,14 +9,14 @@ from .aggregations import Aggregation, _atleast_1d
 from .core import (
     factorize_,
     groupby_reduce,
-    rechunk_for_blockwise,
+    rechunk_for_blockwise as rechunk_array_for_blockwise,
     rechunk_for_cohorts as rechunk_array_for_cohorts,
     reindex_,
 )
-from .xrutils import isnull
+from .xrutils import is_duck_dask_array, isnull
 
 if TYPE_CHECKING:
-    from xarray import DataArray, Dataset, GroupBy, Resample
+    from xarray import DataArray, Dataset, Resample
 
 
 def _get_input_core_dims(group_names, dim, ds, to_group):
@@ -54,11 +53,12 @@ def xarray_reduce(
     func: Union[str, Aggregation],
     expected_groups=None,
     isbin: Union[bool, Sequence[bool]] = False,
+    sort: bool = True,
     dim: Hashable = None,
     split_out: int = 1,
     fill_value=None,
-    method: str = "mapreduce",
-    engine: str = "numpy",
+    method: str = "map-reduce",
+    engine: str = "flox",
     keep_attrs: bool = True,
     skipna: Optional[bool] = None,
     min_count: Optional[int] = None,
@@ -79,17 +79,21 @@ def xarray_reduce(
     isbin : iterable of bool
         If True, corresponding entry in ``expected_groups`` are bin edges.
         If False, the entry in ``expected_groups`` is treated as a simple label.
+    sort : (optional), bool
+        Whether groups should be returned in sorted order. Only applies for dask
+        reductions when ``method`` is not ``"map-reduce"``. For ``"map-reduce"``, the groups
+        are always sorted.
     dim : hashable
         dimension name along which to reduce. If None, reduces across all
         dimensions of `by`
     split_out : int, optional
         Number of output chunks along grouped dimension in output.
-    fill_value :
+    fill_value
         Value used for missing groups in the output i.e. when one of the labels
         in ``expected_groups`` is not actually present in ``by``.
-    method : {"mapreduce", "blockwise", "cohorts"}, optional
+    method : {"map-reduce", "blockwise", "cohorts", "split-reduce"}, optional
         Strategy for reduction of dask arrays only:
-          * ``"mapreduce"``:
+          * ``"map-reduce"``:
             First apply the reduction blockwise on ``array``, then
             combine a few newighbouring blocks, apply the reduction.
             Continue until finalizing. Usually, ``func`` will need
@@ -98,20 +102,34 @@ def xarray_reduce(
           * ``"blockwise"``:
             Only reduce using blockwise and avoid aggregating blocks
             together. Useful for resampling-style reductions where group
-            members are always together. The array is rechunked so that
-            chunk boundaries line up with group boundaries
+            members are always together. If  `by` is 1D,  `array` is automatically
+            rechunked so that chunk boundaries line up with group boundaries
             i.e. each block contains all members of any group present
-            in that block.
+            in that block. For nD `by`, you must make sure that all members of a group
+            are present in a single block.
           * ``"cohorts"``:
             Finds group labels that tend to occur together ("cohorts"),
-            indexes out cohorts and reduces that subset using "mapreduce",
+            indexes out cohorts and reduces that subset using "map-reduce",
             repeat for all cohorts. This works well for many time groupings
             where the group labels repeat at regular intervals like 'hour',
             'month', dayofyear' etc. Optimize chunking ``array`` for this
-            method by first rechunking using ``rechunk_for_cohorts``.
-
-    engine : {"numpy", "numba"}, optional
-        Engine for numpy_groupies
+            method by first rechunking using ``rechunk_for_cohorts``
+            (for 1D ``by`` only).
+          * ``"split-reduce"``:
+            Break out each group into its own array and then ``"map-reduce"``.
+            This is implemented by having each group be its own cohort,
+            and is identical to xarray's default strategy.
+    engine : {"flox", "numpy", "numba"}, optional
+        Algorithm to compute the groupby reduction on non-dask arrays and on each dask chunk:
+          * ``"flox"``:
+            Use an internal implementation where the data is sorted so that
+            all members of a group occur sequentially, and then numpy.ufunc.reduceat
+            is to used for the reduction. This will fall back to ``numpy_groupies.aggregate_numpy``
+            for a reduction that is not yet implemented.
+          * ``"numpy"``:
+            Use the vectorized implementations in ``numpy_groupies.aggregate_numpy``.
+          * ``"numba"``:
+            Use the implementations in ``numpy_groupies.aggregate_numba``.
     keep_attrs : bool, optional
         Preserve attrs?
     skipna : bool, optional
@@ -124,8 +142,8 @@ def xarray_reduce(
         fewer than min_count non-NA values are present the result will be
         NA. Only used if skipna is set to True or defaults to True for the
         array's dtype.
-    finalize_kwargs : dict, optional
-        kwargs passed to the finalize function, like ddof for var, std.
+    **finalize_kwargs :
+        kwargs passed to the finalize function, like ``ddof`` for var, std.
 
     Returns
     -------
@@ -134,7 +152,7 @@ def xarray_reduce(
 
     See Also
     --------
-    dask_groupby.core.groupby_reduce
+    flox.core.groupby_reduce
 
     Raises
     ------
@@ -145,6 +163,10 @@ def xarray_reduce(
     --------
     FIXME: Add docs.
     """
+
+    if skipna is not None and isinstance(func, Aggregation):
+        raise ValueError("skipna must be None when func is an Aggregation.")
+
     for b in by:
         if isinstance(b, xr.DataArray) and b.name is None:
             raise ValueError("Cannot group by unnamed DataArrays.")
@@ -152,6 +174,9 @@ def xarray_reduce(
     if isinstance(isbin, bool):
         isbin = (isbin,) * len(by)
 
+    if not sort:
+        raise NotImplementedError
+
     # eventually  drop the variables we are grouping by
     maybe_drop = [b for b in by if isinstance(b, str)]
     unindexed_dims = tuple(
@@ -162,7 +187,7 @@ def xarray_reduce(
 
     by: Tuple["DataArray"] = tuple(obj[g] if isinstance(g, str) else g for g in by)  # type: ignore
 
-    if len(by) > 1 and any(dask.is_dask_collection(by_) for by_ in by):
+    if len(by) > 1 and any(is_duck_dask_array(by_.data) for by_ in by):
         raise NotImplementedError("Grouping by multiple variables will compute dask variables.")
 
     grouper_dims = set(itertools.chain(*tuple(g.dims for g in by)))
@@ -199,7 +224,7 @@ def xarray_reduce(
         dim = _atleast_1d(dim)
 
     if any(d not in grouper_dims and d not in obj.dims for d in dim):
-        raise ValueError(f"cannot reduce over dimensions {dim}")
+        raise ValueError(f"Cannot reduce over absent dimensions {dim}.")
 
     dims_not_in_groupers = tuple(d for d in dim if d not in grouper_dims)
     if dims_not_in_groupers == dim and not any(isbin):
@@ -230,7 +255,11 @@ def xarray_reduce(
         to_group = xr.DataArray(group_idx, dims=dim, coords={d: by[0][d] for d in by[0].indexes})
     else:
         if expected_groups is None and isinstance(by[0].data, np.ndarray):
-            expected_groups = (np.unique(by[0].data),)
+            uniques = np.unique(by[0].data)
+            nans = isnull(uniques)
+            if nans.any():
+                uniques = uniques[~nans]
+            expected_groups = (uniques,)
         if expected_groups is None:
             raise NotImplementedError(
                 "Please provide expected_groups if not grouping by a numpy-backed DataArray"
@@ -258,7 +287,7 @@ def xarray_reduce(
         if skipna and func in ["all", "any", "count"]:
             raise ValueError(f"skipna cannot be truthy for {func} reductions.")
 
-        if skipna or (skipna is None and array.dtype.kind in "cfO"):
+        if skipna or (skipna is None and isinstance(func, str) and array.dtype.kind in "cfO"):
             if "nan" not in func and func not in ["all", "any", "count"]:
                 func = f"nan{func}"
 
@@ -310,6 +339,7 @@ def xarray_reduce(
         kwargs={
             "func": func,
             "axis": axis,
+            "sort": sort,
             "split_out": split_out,
             "fill_value": fill_value,
             "method": method,
@@ -328,6 +358,12 @@ def xarray_reduce(
         },
     )
 
+    # restore non-dim coord variables without the core dimension
+    # TODO: shouldn't apply_ufunc handle this?
+    for var in set(ds.variables) - set(ds.dims):
+        if all(d not in ds[var].dims for d in dim):
+            actual[var] = ds[var]
+
     for name, expect, isbin_ in zip(group_names, expected_groups, isbin):
         if isbin_:
             expect = [pd.Interval(left, right) for left, right in zip(expect[:-1], expect[1:])]
@@ -368,47 +404,6 @@ def xarray_reduce(
         return actual
 
 
-def xarray_groupby_reduce(
-    groupby: "GroupBy",
-    func: Union[str, Aggregation],
-    split_out: int = 1,
-    method: str = "mapreduce",
-    keep_attrs: bool = True,
-):
-    """Apply on an existing Xarray groupby object for convenience."""
-
-    def wrapper(*args, **kwargs):
-        result, _ = groupby_reduce(*args, **kwargs)
-        return result
-
-    groups = list(groupby.groups.keys())
-    outdim = groupby._unique_coord.name
-    groupdim = groupby._group_dim
-
-    actual = xr.apply_ufunc(
-        wrapper,
-        groupby._obj,
-        groupby._group,
-        input_core_dims=[[groupdim], [groupdim]],
-        # for xarray's test_groupby_duplicate_coordinate_labels
-        exclude_dims=set(groupdim),
-        output_core_dims=[[outdim]],
-        dask="allowed",
-        dask_gufunc_kwargs=dict(output_sizes={outdim: len(groups)}),
-        keep_attrs=keep_attrs,
-        kwargs={
-            "func": func,
-            "axis": -1,
-            "split_out": split_out,
-            "expected_groups": groups,
-            "method": method,
-        },
-    )
-    actual[outdim] = groups
-
-    return actual
-
-
 def rechunk_for_cohorts(
     obj: Union["DataArray", "Dataset"],
     dim: str,
@@ -421,26 +416,27 @@ def rechunk_for_cohorts(
 
     Parameters
     ----------
-    array: DataArray or Dataset
+    obj : DataArray or Dataset
         array to rechunk
-    dim: str
+    dim : str
         Dimension to rechunk
-    labels: DataArray
+    labels : DataArray
         1D Group labels to align chunks with. This routine works
         well when ``labels`` has repeating patterns: e.g.
         ``1, 2, 3, 1, 2, 3, 4, 1, 2, 3`` though there is no requirement
         that the pattern must contain sequences.
-    force_new_chunk_at:
-        label at which we always start a new chunk. For
-        the example ``labels`` array, this would be `1``.
-    chunksize: int, optional
+    force_new_chunk_at : Sequence
+        Labels at which we always start a new chunk. For
+        the example ``labels`` array, this would be `1`.
+    chunksize : int, optional
         nominal chunk size. Chunk size is exceded when the label
         in ``force_new_chunk_at`` is less than ``chunksize//2`` elements away.
         If None, uses median chunksize along ``dim``.
+
     Returns
     -------
-    dask.array.Array
-        rechunked array
+    DataArray or Dataset
+        Xarray object with rechunked arrays.
     """
     return _rechunk(
         rechunk_array_for_cohorts,
@@ -452,16 +448,30 @@ def rechunk_for_cohorts(
     )
 
 
-def rechunk_to_group_boundaries(obj: Union["DataArray", "Dataset"], dim: str, labels: "DataArray"):
+def rechunk_for_blockwise(obj: Union["DataArray", "Dataset"], dim: str, labels: "DataArray"):
     """
     Rechunks array so that group boundaries line up with chunk boundaries, allowing
-    parallel group reductions.
+    embarassingly parallel group reductions.
 
-    This only works when the groups are sequential (e.g. labels = [0,0,0,1,1,1,1,2,2]).
+    This only works when the groups are sequential
+    (e.g. labels = ``[0,0,0,1,1,1,1,2,2]``).
     Such patterns occur when using ``.resample``.
-    """
 
-    return _rechunk(rechunk_for_blockwise, obj, dim, labels)
+    Parameters
+    ----------
+    obj : DataArray or Dataset
+        Array to rechunk
+    dim : hashable
+        Name of dimension to rechunk
+    labels : DataArray
+        Group labels
+
+    Returns
+    -------
+    DataArray or Dataset
+        Xarray object with rechunked arrays.
+    """
+    return _rechunk(rechunk_array_for_blockwise, obj, dim, labels)
 
 
 def _rechunk(func, obj, dim, labels, **kwargs):
@@ -489,6 +499,7 @@ def resample_reduce(
     resampler: "Resample",
     func: Union[str, Aggregation],
     keep_attrs: bool = True,
+    **kwargs,
 ):
 
     obj = resampler._obj
@@ -512,6 +523,7 @@ def resample_reduce(
             method="blockwise",
             expected_groups=(resampler._unique_coord.data,),
             keep_attrs=keep_attrs,
+            **kwargs,
         )
         .rename({"__resample_dim__": dim})
         .transpose(dim, ...)
