@@ -1,8 +1,5 @@
-import dask
-import dask.array as da
 import numpy as np
 import pytest
-from dask.array import from_array
 from numpy_groupies.aggregate_numpy import aggregate
 
 from flox.core import (
@@ -13,14 +10,27 @@ from flox.core import (
     reindex_,
 )
 
-from . import assert_equal, raise_if_dask_computes
+from . import assert_equal, engine, has_dask, raise_if_dask_computes, requires_dask
 
 labels = np.array([0, 0, 2, 2, 2, 1, 1, 2, 2, 1, 1, 0])
 nan_labels = labels.astype(float)  # copy
 nan_labels[:5] = np.nan
 labels2d = np.array([labels[:5], np.flip(labels[:5])])
 
-dask.config.set(scheduler="sync")
+# isort:off
+if has_dask:
+    import dask
+    import dask.array as da
+    from dask.array import from_array
+
+    dask.config.set(scheduler="sync")
+else:
+
+    def dask_array_ones(*args):
+        return None
+
+
+# isort:on
 
 ALL_FUNCS = (
     "sum",
@@ -56,7 +66,6 @@ def test_alignment_error():
         groupby_reduce(da, labels, func="mean")
 
 
-@pytest.mark.parametrize("engine", ["numpy", "numba"])
 @pytest.mark.parametrize("dtype", (float, int))
 @pytest.mark.parametrize("chunk, split_out", [(False, 1), (True, 1), (True, 2), (True, 3)])
 @pytest.mark.parametrize("expected_groups", [None, [0, 1, 2], np.array([0, 1, 2])])
@@ -90,7 +99,7 @@ def test_groupby_reduce(
 ):
     array = array.astype(dtype)
     if chunk:
-        if expected_groups is None:
+        if not has_dask or expected_groups is None:
             pytest.skip()
         array = da.from_array(array, chunks=(3,) if array.ndim == 1 else (1, 3))
         by = da.from_array(by, chunks=(3,) if by.ndim == 1 else (1, 3))
@@ -114,13 +123,13 @@ def test_groupby_reduce(
     assert_equal(expected, result)
 
 
-@pytest.mark.parametrize("engine", ["numpy", "numba"])
 @pytest.mark.parametrize("size", ((12,), (12, 5)))
 @pytest.mark.parametrize("func", ALL_FUNCS)
 def test_groupby_reduce_all(size, func, engine):
 
     by = np.ones(size[-1])
-    array = np.random.randn(*size)
+    rng = np.random.default_rng(12345)
+    array = rng.random(size)
     if "nan" in func and "nanarg" not in func:
         array[[1, 4, 5], ...] = np.nan
     elif "nanarg" in func and len(size) > 1:
@@ -143,7 +152,9 @@ def test_groupby_reduce_all(size, func, engine):
             assert actual.dtype.kind == "i"
         assert_equal(actual, expected)
 
-        for method in ["mapreduce", "cohorts"]:
+        if not has_dask:
+            continue
+        for method in ["map-reduce", "cohorts", "split-reduce"]:
             actual, _ = groupby_reduce(
                 da.from_array(array, chunks=3),
                 by,
@@ -157,12 +168,14 @@ def test_groupby_reduce_all(size, func, engine):
             assert_equal(actual, expected)
 
 
+@requires_dask
 @pytest.mark.parametrize("size", ((12,), (12, 5)))
 @pytest.mark.parametrize("func", ("argmax", "nanargmax", "argmin", "nanargmin"))
 def test_arg_reduction_dtype_is_int(size, func):
     """avoid bugs being hidden by the xfail in the above test."""
 
-    array = np.random.randn(*size)
+    rng = np.random.default_rng(12345)
+    array = rng.random(size)
     by = np.ones(size[-1])
 
     if "nanarg" in func and len(size) > 1:
@@ -171,10 +184,10 @@ def test_arg_reduction_dtype_is_int(size, func):
     expected = getattr(np, func)(array, axis=-1)
     expected = np.expand_dims(expected, -1)
 
-    actual, _ = groupby_reduce(array, by, func=func)
+    actual, _ = groupby_reduce(array, by, func=func, engine="numpy")
     assert actual.dtype.kind == "i"
 
-    actual, _ = groupby_reduce(da.from_array(array, chunks=3), by, func=func)
+    actual, _ = groupby_reduce(da.from_array(array, chunks=3), by, func=func, engine="numpy")
     assert actual.dtype.kind == "i"
 
 
@@ -185,6 +198,17 @@ def test_groupby_reduce_count():
     assert_equal(result, [1, 1, 2])
 
 
+def test_func_is_aggregation():
+    from flox.aggregations import mean
+
+    array = np.array([0, 0, np.nan, np.nan, np.nan, 1, 1])
+    labels = np.array(["a", "b", "b", "b", "c", "c", "c"])
+    expected, _ = groupby_reduce(array, labels, func="mean")
+    actual, _ = groupby_reduce(array, labels, func=mean)
+    assert_equal(actual, expected)
+
+
+@requires_dask
 @pytest.mark.parametrize("func", ("sum", "prod"))
 @pytest.mark.parametrize("dtype", [np.float32, np.float64, np.int32, np.int64])
 def test_groupby_reduce_preserves_dtype(dtype, func):
@@ -219,43 +243,33 @@ def test_numpy_reduce_nd_md():
     assert_equal(expected, actual)
 
 
-@pytest.mark.parametrize(
-    "func",
-    (
-        "sum",
-        "count",
-        "prod",
-        "mean",
-        "std",
-        "var",
-        "max",
-        "min",
-        "argmax",
-        "argmin",
-        # "first",
-        # "last",
-    ),
-)
+@requires_dask
+@pytest.mark.parametrize("func", ALL_FUNCS)
 @pytest.mark.parametrize("add_nan", [False, True])
 @pytest.mark.parametrize("dtype", (float,))
 @pytest.mark.parametrize(
-    "array, group_chunks",
+    "shape, array_chunks, group_chunks",
     [
-        (da.ones((12,), (3,)), 3),  # form 1
-        (da.ones((12,), (3,)), (4,)),  # form 1, chunks not aligned
-        (da.ones((12,), ((3, 5, 4),)), (2,)),  # form 1
-        (da.ones((10, 12), (3, 3)), -1),  # form 3
-        (da.ones((10, 12), (3, 3)), 3),  # form 3
+        ((12,), (3,), 3),  # form 1
+        ((12,), (3,), (4,)),  # form 1, chunks not aligned
+        ((12,), ((3, 5, 4),), (2,)),  # form 1
+        ((10, 12), (3, 3), -1),  # form 3
+        ((10, 12), (3, 3), 3),  # form 3
     ],
 )
-def test_groupby_agg_dask(func, array, group_chunks, add_nan, dtype):
+def test_groupby_agg_dask(func, shape, array_chunks, group_chunks, add_nan, dtype, engine):
     """Tests groupby_reduce with dask arrays against groupby_reduce with numpy arrays"""
 
-    array = array.astype(dtype)
+    rng = np.random.default_rng(12345)
+    array = dask.array.from_array(rng.random(shape), chunks=array_chunks).astype(dtype)
+    array = dask.array.ones(shape, chunks=array_chunks)
 
     if func in ["first", "last"]:
         pytest.skip()
 
+    if "arg" in func and engine == "flox":
+        pytest.skip()
+
     labels = np.array([0, 0, 2, 2, 2, 1, 1, 2, 2, 1, 1, 0])
     if add_nan:
         labels = labels.astype(float)
@@ -264,50 +278,58 @@ def test_groupby_agg_dask(func, array, group_chunks, add_nan, dtype):
 
     kwargs = dict(func=func, expected_groups=[0, 1, 2], fill_value=123)
 
-    by = from_array(labels, group_chunks)
-    expected, _ = groupby_reduce(array.compute(), by.compute(), **kwargs)
+    expected, _ = groupby_reduce(array.compute(), labels, engine="numpy", **kwargs)
+    actual, _ = groupby_reduce(array.compute(), labels, engine=engine, **kwargs)
+    assert_equal(actual, expected)
+
     with raise_if_dask_computes():
-        actual, _ = groupby_reduce(array, by, **kwargs)
+        actual, _ = groupby_reduce(array, labels, engine=engine, **kwargs)
+    assert_equal(actual, expected)
+
+    by = from_array(labels, group_chunks)
+    with raise_if_dask_computes():
+        actual, _ = groupby_reduce(array, by, engine=engine, **kwargs)
     assert_equal(expected, actual)
 
     kwargs["expected_groups"] = [0, 2, 1]
     with raise_if_dask_computes():
-        actual, groups = groupby_reduce(array, by, **kwargs)
+        actual, groups = groupby_reduce(array, by, engine=engine, **kwargs)
     assert_equal(groups, [0, 2, 1])
     assert_equal(expected, actual[..., [0, 2, 1]])
 
 
-def test_numpy_reduce_axis_subset():
+def test_numpy_reduce_axis_subset(engine):
     # TODO: add NaNs
     by = labels2d
     array = np.ones_like(by)
-    result, _ = groupby_reduce(array, by, "count", axis=1)
+    result, _ = groupby_reduce(array, by, "count", axis=1, engine=engine)
     assert_equal(result, [[2, 3], [2, 3]])
 
     by = np.broadcast_to(labels2d, (3, *labels2d.shape))
     array = np.ones_like(by)
-    result, _ = groupby_reduce(array, by, "count", axis=1)
+    result, _ = groupby_reduce(array, by, "count", axis=1, engine=engine)
     subarr = np.array([[1, 1], [1, 1], [0, 2], [1, 1], [1, 1]])
     expected = np.tile(subarr, (3, 1, 1))
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by, "count", axis=2)
+    result, _ = groupby_reduce(array, by, "count", axis=2, engine=engine)
     subarr = np.array([[2, 3], [2, 3]])
     expected = np.tile(subarr, (3, 1, 1))
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by, "count", axis=(1, 2))
+    result, _ = groupby_reduce(array, by, "count", axis=(1, 2), engine=engine)
     expected = np.array([[4, 6], [4, 6], [4, 6]])
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by, "count", axis=(2, 1))
+    result, _ = groupby_reduce(array, by, "count", axis=(2, 1), engine=engine)
     assert_equal(result, expected)
 
-    result, _ = groupby_reduce(array, by[0, ...], "count", axis=(1, 2))
+    result, _ = groupby_reduce(array, by[0, ...], "count", axis=(1, 2), engine=engine)
     expected = np.array([[4, 6], [4, 6], [4, 6]])
     assert_equal(result, expected)
 
 
+@requires_dask
 def test_dask_reduce_axis_subset():
 
     by = labels2d
@@ -358,12 +380,15 @@ def test_dask_reduce_axis_subset():
         )
 
 
+@requires_dask
 @pytest.mark.parametrize("func", ALL_FUNCS)
-@pytest.mark.parametrize("engine", ["numpy", "numba"])
 @pytest.mark.parametrize(
     "axis", [None, (0, 1, 2), (0, 1), (0, 2), (1, 2), 0, 1, 2, (0,), (1,), (2,)]
 )
 def test_groupby_reduce_axis_subset_against_numpy(func, axis, engine):
+    if "arg" in func and engine == "flox":
+        pytest.skip()
+
     if not isinstance(axis, int) and "arg" in func and (axis is None or len(axis) > 1):
         pytest.skip()
     if func in ["all", "any"]:
@@ -372,7 +397,8 @@ def test_groupby_reduce_axis_subset_against_numpy(func, axis, engine):
         fill_value = 123
     # tests against the numpy output to make sure dask compute matches
     by = np.broadcast_to(labels2d, (3, *labels2d.shape))
-    array = np.ones_like(by)
+    rng = np.random.default_rng(12345)
+    array = rng.random(by.shape)
     kwargs = dict(
         func=func, axis=axis, expected_groups=[0, 2], fill_value=fill_value, engine=engine
     )
@@ -383,10 +409,13 @@ def test_groupby_reduce_axis_subset_against_numpy(func, axis, engine):
             **kwargs,
         )
     expected, _ = groupby_reduce(array, by, **kwargs)
+    if engine == "flox":
+        kwargs.pop("engine")
+        expected_npg, _ = groupby_reduce(array, by, **kwargs, engine="numpy")
+        assert_equal(expected_npg, expected)
     assert_equal(actual, expected)
 
 
-@pytest.mark.parametrize("engine", ["numpy", "numba"])
 @pytest.mark.parametrize("chunks", [None, (2, 2, 3)])
 @pytest.mark.parametrize(
     "axis, groups, expected_shape",
@@ -399,6 +428,8 @@ def test_groupby_reduce_axis_subset_against_numpy(func, axis, engine):
 def test_groupby_reduce_nans(chunks, axis, groups, expected_shape, engine):
     def _maybe_chunk(arr):
         if chunks:
+            if not has_dask:
+                pytest.skip()
             return da.from_array(arr, chunks=chunks)
         else:
             return arr
@@ -428,7 +459,7 @@ def test_groupby_reduce_nans(chunks, axis, groups, expected_shape, engine):
     # by = np.broadcast_to(labels2d, (3, *labels2d.shape))
 
 
-@pytest.mark.parametrize("engine", ["numpy", "numba"])
+@requires_dask
 def test_groupby_all_nan_blocks(engine):
     labels = np.array([0, 0, 2, 2, 2, 1, 1, 2, 2, 1, 1, 0])
     nan_labels = labels.astype(float)  # copy
@@ -486,14 +517,18 @@ def test_npg_nanarg_bug(func):
     assert_equal(actual, expected)
 
 
-@pytest.mark.parametrize("chunks", ((), (2,)))
-def test_groupby_bins(chunks):
-    array = [1, 1, 1, 1, 1]
-    labels = [1, 1.5, 1.9, 2, 3]
+@pytest.mark.parametrize("chunk_labels", [False, True])
+@pytest.mark.parametrize("chunks", ((), (1,), (2,)))
+def test_groupby_bins(chunk_labels, chunks, engine) -> None:
+    array = [1, 1, 1, 1, 1, 1]
+    labels = [0.2, 1.5, 1.9, 2, 3, 20]
 
     if chunks:
+        if not has_dask:
+            pytest.skip()
         array = dask.array.from_array(array, chunks=chunks)
-        labels = dask.array.from_array(labels, chunks=chunks)
+        if chunk_labels:
+            labels = dask.array.from_array(labels, chunks=chunks)
 
     with raise_if_dask_computes():
         actual, groups = groupby_reduce(
@@ -503,6 +538,7 @@ def test_groupby_bins(chunks):
             expected_groups=np.array([1, 2, 4, 5]),
             isbin=True,
             fill_value=0,
+            engine=engine,
         )
     expected = np.array([3, 1, 0])
     assert_equal(groups, np.array([0, 1, 2]))
@@ -531,22 +567,34 @@ def test_rechunk_for_blockwise(inchunks, expected):
     assert _get_optimal_chunks_for_groups(inchunks, labels) == expected
 
 
+@requires_dask
 @pytest.mark.parametrize(
     "expected, labels, chunks, merge",
     [
         [[[1, 2, 3, 4]], [1, 2, 3, 1, 2, 3, 4], (3, 4), True],
         [[[1, 2, 3], [4]], [1, 2, 3, 1, 2, 3, 4], (3, 4), False],
         [[[1], [2], [3], [4]], [1, 2, 3, 1, 2, 3, 4], (2, 2, 2, 1), False],
-        [[[1], [2], [3], [4]], [1, 2, 3, 1, 2, 3, 4], (2, 2, 2, 1), True],
+        [[[3], [2], [1], [4]], [1, 2, 3, 1, 2, 3, 4], (2, 2, 2, 1), True],
         [[[1, 2, 3], [4]], [1, 2, 3, 1, 2, 3, 4], (3, 3, 1), True],
         [[[1, 2, 3], [4]], [1, 2, 3, 1, 2, 3, 4], (3, 3, 1), False],
+        [
+            [[2, 3, 4, 1], [5], [0]],
+            np.repeat(np.arange(6), [4, 4, 12, 2, 3, 4]),
+            (4, 8, 4, 9, 4),
+            True,
+        ],
     ],
 )
 def test_find_group_cohorts(expected, labels, chunks, merge):
-    actual = list(find_group_cohorts(labels, chunks, merge))
+    actual = list(find_group_cohorts(labels, (chunks,), merge, method="cohorts"))
+    assert actual == expected, (actual, expected)
+
+    actual = find_group_cohorts(labels, (chunks,), merge, method="split-reduce")
+    expected = [[label] for label in np.unique(labels)]
     assert actual == expected, (actual, expected)
 
 
+@requires_dask
 @pytest.mark.parametrize(
     "chunk_at,expected",
     [
@@ -562,8 +610,11 @@ def test_rechunk_for_cohorts(chunk_at, expected):
     assert rechunked.chunks == expected
 
 
+@pytest.mark.parametrize("fill_value", [123, np.nan])
 @pytest.mark.parametrize("func", ALL_FUNCS)
-def test_fill_value_behaviour(func):
+def test_fill_value_behaviour(func, fill_value, engine):
+    # fill_value = np.nan tests promotion of int counts to float
+    # This is used by xarray
     if func in ["all", "any"] or "arg" in func:
         pytest.skip()
 
@@ -576,13 +627,116 @@ def test_fill_value_behaviour(func):
     else:
         npfunc = getattr(np, func)
 
-    fill_value = 123
     by = np.array([1, 2, 3, 1, 2, 3])
     array = np.array([np.nan, 1, 1, np.nan, 1, 1])
     actual, _ = groupby_reduce(
-        array, by, func=func, fill_value=fill_value, expected_groups=[0, 1, 2, 3]
+        array, by, func=func, engine=engine, fill_value=fill_value, expected_groups=[0, 1, 2, 3]
     )
     expected = np.array(
         [fill_value, npfunc([np.nan, np.nan]), npfunc([1.0, 1.0]), npfunc([1.0, 1.0])]
     )
     assert_equal(actual, expected)
+
+
+@requires_dask
+@pytest.mark.parametrize("func", ["mean", "sum"])
+@pytest.mark.parametrize("dtype", ["float32", "float64", "int32", "int64"])
+def test_dtype_preservation(dtype, func, engine):
+    if func == "sum" or (func == "mean" and "float" in dtype):
+        expected = np.dtype(dtype)
+    elif func == "mean" and "int" in dtype:
+        expected = np.float64
+    array = np.ones((20,), dtype=dtype)
+    by = np.ones(array.shape, dtype=int)
+    actual, _ = groupby_reduce(array, by, func=func, engine=engine)
+    assert actual.dtype == expected
+
+    array = dask.array.from_array(array, chunks=(4,))
+    actual, _ = groupby_reduce(array, by, func=func, engine=engine)
+    assert actual.dtype == expected
+
+
+@requires_dask
+@pytest.mark.parametrize("method", ["split-reduce", "map-reduce", "cohorts"])
+def test_cohorts(method):
+    repeats = [4, 4, 12, 2, 3, 4]
+    labels = np.repeat(np.arange(6), repeats)
+    array = dask.array.from_array(labels, chunks=(4, 8, 4, 9, 4))
+
+    actual, actual_groups = groupby_reduce(array, labels, func="count", method=method)
+    assert_equal(actual_groups, np.arange(6))
+    assert_equal(actual, repeats)
+
+
+@requires_dask
+@pytest.mark.parametrize("func", ALL_FUNCS)
+@pytest.mark.parametrize("axis", (-1, None))
+@pytest.mark.parametrize("method", ["blockwise", "cohorts", "map-reduce", "split-reduce"])
+def test_cohorts_nd_by(func, method, axis, engine):
+    o = dask.array.ones((3,), chunks=-1)
+    o2 = dask.array.ones((2, 3), chunks=-1)
+
+    array = dask.array.block([[o, 2 * o], [3 * o2, 4 * o2]])
+    by = array.compute().astype(int)
+    by[0, 1] = 30
+    by[2, 1] = 40
+    by[0, 4] = 31
+    array = np.broadcast_to(array, (2, 3) + array.shape)
+
+    if "arg" in func and (axis is None or engine == "flox"):
+        pytest.skip()
+
+    if func in ["any", "all"]:
+        fill_value = False
+    else:
+        fill_value = -123
+
+    if axis is not None and method != "map-reduce":
+        pytest.xfail()
+
+    kwargs = dict(func=func, engine=engine, method=method, axis=axis, fill_value=fill_value)
+    actual, _ = groupby_reduce(array, by, **kwargs)
+    expected, sorted_groups = groupby_reduce(array.compute(), by, **kwargs)
+    assert_equal(actual, expected)
+
+    actual, groups = groupby_reduce(array, by, sort=False, **kwargs)
+    if method == "cohorts":
+        assert_equal(groups, [4, 3, 40, 2, 31, 1, 30])
+    elif method == "blockwise":
+        assert_equal(groups, [1, 30, 2, 31, 3, 40, 4])
+    else:
+        assert_equal(groups, [1, 2, 3, 4, 30, 31, 40])
+    reindexed = reindex_(actual, groups, sorted_groups)
+    assert_equal(reindexed, expected)
+
+
+@pytest.mark.parametrize("func", ["sum", "count"])
+@pytest.mark.parametrize(
+    "fill_value, expected", ((None, np.floating), (0, np.integer), (np.nan, np.floating))
+)
+def test_dtype_promotion(func, fill_value, expected, engine):
+    array = np.array([1, 1])
+    by = [0, 1]
+
+    actual, _ = groupby_reduce(
+        array, by, func=func, expected_groups=[1, 2], fill_value=fill_value, engine=engine
+    )
+    assert np.issubdtype(actual.dtype, expected)
+
+
+@pytest.mark.parametrize("func", ["mean", "nanmean"])
+def test_empty_bins(func, engine):
+    array = np.ones((2, 3, 2))
+    by = np.broadcast_to([0, 1], array.shape)
+
+    actual, _ = groupby_reduce(
+        array,
+        by,
+        func=func,
+        expected_groups=[-1, 0, 1, 2],
+        isbin=True,
+        engine=engine,
+        axis=(0, 1, 2),
+    )
+    expected = np.array([1.0, 1.0, np.nan])
+    assert_equal(actual, expected)
