@@ -1,28 +1,29 @@
-import dask
 import numpy as np
 import pandas as pd
 import pytest
-import xarray as xr
 
-from flox.xarray import (
-    rechunk_to_group_boundaries,
-    resample_reduce,
-    xarray_groupby_reduce,
-    xarray_reduce,
-)
+# isort: off
+xr = pytest.importorskip("xarray")
+# isort: on
 
-from . import assert_equal, raise_if_dask_computes
+from flox.xarray import rechunk_for_blockwise, resample_reduce, xarray_reduce
 
-dask.config.set(scheduler="sync")
+from . import assert_equal, engine, has_dask, raise_if_dask_computes, requires_dask
+
+# isort: off
+if has_dask:
+    import dask
+
+    dask.config.set(scheduler="sync")
+# isort: on
 
 try:
     # Should test against legacy xarray implementation
-    xr.set_options(use_numpy_groupies=False)
+    xr.set_options(use_flox=False)
 except ValueError:
     pass
 
 
-@pytest.mark.parametrize("engine", ["numpy", "numba"])
 @pytest.mark.parametrize("min_count", [None, 1, 3])
 @pytest.mark.parametrize("add_nan", [True, False])
 @pytest.mark.parametrize("skipna", [True, False])
@@ -50,6 +51,14 @@ def test_xarray_reduce(skipna, add_nan, min_count, engine):
     )
     assert_equal(expected, actual)
 
+    da["labels2"] = da.labels2.astype(float)
+    da["labels2"][0] = np.nan
+    expected = da.groupby("labels2").sum(skipna=skipna, min_count=min_count)
+    actual = xarray_reduce(
+        da, "labels2", func="sum", skipna=skipna, min_count=min_count, engine=engine
+    )
+    assert_equal(expected, actual)
+
     # test dimension ordering
     # actual = xarray_reduce(
     #    da.transpose("y", ...), "labels", func="sum", skipna=skipna, min_count=min_count
@@ -57,33 +66,7 @@ def test_xarray_reduce(skipna, add_nan, min_count, engine):
     # assert_equal(expected, actual)
 
 
-def test_xarray_groupby_reduce():
-    arr = np.ones((4, 12))
-
-    labels = np.array(["a", "a", "c", "c", "c", "b", "b", "c", "c", "b", "b", "f"])
-    labels = np.array(labels)
-    labels2 = np.array([1, 2, 2, 1])
-
-    da = xr.DataArray(
-        arr, dims=("x", "y"), coords={"labels2": ("x", labels2), "labels": ("y", labels)}
-    ).expand_dims(z=4)
-
-    grouped = da.groupby("labels")
-    expected = grouped.mean()
-    actual = xarray_groupby_reduce(grouped, "mean")
-    assert_equal(expected, actual)
-
-    actual = xarray_groupby_reduce(da.transpose("y", ...).groupby("labels"), "mean")
-    assert_equal(expected, actual)
-
-    # TODO: fails because of stacking
-    # grouped = da.groupby("labels2")
-    # expected = grouped.mean()
-    # actual = xarray_groupby_reduce(grouped, "mean")
-    # assert_equal(expected, actual)
-
-
-def test_xarray_reduce_multiple_groupers():
+def test_xarray_reduce_multiple_groupers(engine):
     arr = np.ones((4, 12))
 
     labels = np.array(["a", "a", "c", "c", "c", "b", "b", "c", "c", "b", "b", "f"])
@@ -100,53 +83,57 @@ def test_xarray_reduce_multiple_groupers():
         coords={"labels": ["a", "c", "b", "f"], "labels2": [1, 2]},
     ).expand_dims(z=4)
 
-    actual = xarray_reduce(da, da.labels, da.labels2, func="count")
+    actual = xarray_reduce(da, da.labels, da.labels2, func="count", engine=engine)
     xr.testing.assert_identical(expected, actual)
 
-    actual = xarray_reduce(da, "labels", da.labels2, func="count")
+    actual = xarray_reduce(da, "labels", da.labels2, func="count", engine=engine)
     xr.testing.assert_identical(expected, actual)
 
-    actual = xarray_reduce(da, "labels", "labels2", func="count")
+    actual = xarray_reduce(da, "labels", "labels2", func="count", engine=engine)
     xr.testing.assert_identical(expected, actual)
 
-    with raise_if_dask_computes():
-        actual = xarray_reduce(da.chunk({"x": 2, "z": 1}), da.labels, da.labels2, func="count")
-    xr.testing.assert_identical(expected, actual)
+    if has_dask:
+        with raise_if_dask_computes():
+            actual = xarray_reduce(
+                da.chunk({"x": 2, "z": 1}), da.labels, da.labels2, func="count", engine=engine
+            )
+        xr.testing.assert_identical(expected, actual)
 
-    with pytest.raises(NotImplementedError):
-        actual = xarray_reduce(da.chunk({"x": 2, "z": 1}), "labels", "labels2", func="count")
+        with pytest.raises(NotImplementedError):
+            xarray_reduce(da.chunk({"x": 2, "z": 1}), "labels", "labels2", func="count")
     # xr.testing.assert_identical(expected, actual)
 
 
-def test_xarray_reduce_single_grouper():
+@requires_dask
+def test_xarray_reduce_single_grouper(engine):
 
     # DataArray
     ds = xr.tutorial.open_dataset("rasm", chunks={"time": 9})
-    actual = xarray_reduce(ds.Tair, ds.time.dt.month, func="mean")
+    actual = xarray_reduce(ds.Tair, ds.time.dt.month, func="mean", engine=engine)
     expected = ds.Tair.groupby("time.month").mean()
     xr.testing.assert_allclose(actual, expected)
 
     # Ellipsis reduction
-    actual = xarray_reduce(ds.Tair, ds.time.dt.month, func="mean", dim=...)
+    actual = xarray_reduce(ds.Tair, ds.time.dt.month, func="mean", dim=..., engine=engine)
     expected = ds.Tair.groupby("time.month").mean(...)
     xr.testing.assert_allclose(actual, expected)
 
     # Dataset
     expected = ds.groupby("time.month").mean()
-    actual = xarray_reduce(ds, ds.time.dt.month, func="mean")
+    actual = xarray_reduce(ds, ds.time.dt.month, func="mean", engine=engine)
     xr.testing.assert_allclose(actual, expected)
 
     # add data var with missing grouper dim
     ds["foo"] = ("bar", [1, 2, 3])
     expected = ds.groupby("time.month").mean()
-    actual = xarray_reduce(ds, ds.time.dt.month, func="mean")
+    actual = xarray_reduce(ds, ds.time.dt.month, func="mean", engine=engine)
     xr.testing.assert_allclose(actual, expected)
     del ds["foo"]
 
     # non-dim coord with missing grouper dim
     ds.coords["foo"] = ("bar", [1, 2, 3])
     expected = ds.groupby("time.month").mean()
-    actual = xarray_reduce(ds, ds.time.dt.month, func="mean")
+    actual = xarray_reduce(ds, ds.time.dt.month, func="mean", engine=engine)
     xr.testing.assert_allclose(actual, expected)
     del ds["foo"]
 
@@ -167,31 +154,36 @@ def test_xarray_reduce_errors():
         xarray_reduce(da, by, func="mean")
 
     by.name = "by"
-    with pytest.raises(ValueError, match="cannot reduce over"):
+    with pytest.raises(ValueError, match="Cannot reduce over"):
         xarray_reduce(da, by, func="mean", dim="foo")
 
-    with pytest.raises(NotImplementedError, match="provide expected_groups"):
-        xarray_reduce(da, by.chunk(), func="mean")
+    if has_dask:
+        with pytest.raises(NotImplementedError, match="provide expected_groups"):
+            xarray_reduce(da, by.chunk(), func="mean")
 
 
 @pytest.mark.parametrize("isdask", [True, False])
 @pytest.mark.parametrize("dataarray", [True, False])
 @pytest.mark.parametrize("chunklen", [27, 4 * 31 + 1, 4 * 31 + 20])
-def test_xarray_resample(chunklen, isdask, dataarray):
-    ds = xr.tutorial.open_dataset("air_temperature", chunks={"time": chunklen})
-    if not isdask:
-        ds = ds.compute()
+def test_xarray_resample(chunklen, isdask, dataarray, engine):
+    if isdask:
+        if not has_dask:
+            pytest.skip()
+        ds = xr.tutorial.open_dataset("air_temperature", chunks={"time": chunklen})
+    else:
+        ds = xr.tutorial.open_dataset("air_temperature")
 
     if dataarray:
         ds = ds.air
 
     resampler = ds.resample(time="M")
-    actual = resample_reduce(resampler, "mean")
+    actual = resample_reduce(resampler, "mean", engine=engine)
     expected = resampler.mean()
     xr.testing.assert_allclose(actual, expected)
 
 
-def test_xarray_resample_dataset_multiple_arrays():
+@requires_dask
+def test_xarray_resample_dataset_multiple_arrays(engine):
     # regression test for #35
     times = pd.date_range("2000", periods=5)
     foo = xr.DataArray(range(5), dims=["time"], coords=[times], name="foo")
@@ -202,10 +194,11 @@ def test_xarray_resample_dataset_multiple_arrays():
     # The separate computes are necessary here to force xarray
     # to compute all variables in result at the same time.
     expected = resampler.mean().compute()
-    result = resample_reduce(resampler, "mean").compute()
+    result = resample_reduce(resampler, "mean", engine=engine).compute()
     xr.testing.assert_allclose(expected, result)
 
 
+@requires_dask
 @pytest.mark.parametrize(
     "inchunks, expected",
     [
@@ -223,19 +216,19 @@ def test_xarray_resample_dataset_multiple_arrays():
         [(10,), (10,)],
     ],
 )
-def test_rechunk_to_group_boundaries(inchunks, expected):
+def test_rechunk_for_blockwise(inchunks, expected):
     labels = np.array([1, 1, 1, 2, 2, 3, 3, 5, 5, 5])
 
     da = xr.DataArray(dask.array.ones((10,), chunks=inchunks), dims="x", name="foo")
-    rechunked = rechunk_to_group_boundaries(da, "x", xr.DataArray(labels, dims="x"))
+    rechunked = rechunk_for_blockwise(da, "x", xr.DataArray(labels, dims="x"))
     assert rechunked.chunks == (expected,)
 
     da = xr.DataArray(dask.array.ones((5, 10), chunks=(-1, inchunks)), dims=("y", "x"), name="foo")
-    rechunked = rechunk_to_group_boundaries(da, "x", xr.DataArray(labels, dims="x"))
+    rechunked = rechunk_for_blockwise(da, "x", xr.DataArray(labels, dims="x"))
     assert rechunked.chunks == ((5,), expected)
-
     ds = da.to_dataset()
-    rechunked = rechunk_to_group_boundaries(ds, "x", xr.DataArray(labels, dims="x"))
+
+    rechunked = rechunk_for_blockwise(ds, "x", xr.DataArray(labels, dims="x"))
     assert rechunked.foo.chunks == ((5,), expected)
 
 
@@ -244,15 +237,15 @@ def test_rechunk_to_group_boundaries(inchunks, expected):
 # TODO: dim=None, dim=Ellipsis, groupby unindexed dim
 
 
-def test_groupby_duplicate_coordinate_labels():
+def test_groupby_duplicate_coordinate_labels(engine):
     # fix for http://stackoverflow.com/questions/38065129
     array = xr.DataArray([1, 2, 3], [("x", [1, 1, 2])])
     expected = xr.DataArray([3, 3], [("x", [1, 2])])
-    actual = xarray_reduce(array, array.x, func="sum")
+    actual = xarray_reduce(array, array.x, func="sum", engine=engine)
     assert_equal(expected, actual)
 
 
-def test_multi_index_groupby_sum():
+def test_multi_index_groupby_sum(engine):
     # regression test for xarray GH873
     ds = xr.Dataset(
         {"foo": (("x", "y", "z"), np.ones((3, 4, 2)))},
@@ -260,16 +253,18 @@ def test_multi_index_groupby_sum():
     )
     expected = ds.sum("z")
     stacked = ds.stack(space=["x", "y"])
-    actual = xarray_reduce(stacked, "space", dim="z", func="sum")
+    actual = xarray_reduce(stacked, "space", dim="z", func="sum", engine=engine)
     assert_equal(expected, actual.unstack("space"))
 
 
 @pytest.mark.parametrize("chunks", (None, 2))
-def test_xarray_groupby_bins(chunks):
+def test_xarray_groupby_bins(chunks, engine):
     array = xr.DataArray([1, 1, 1, 1, 1], dims="x")
     labels = xr.DataArray([1, 1.5, 1.9, 2, 3], dims="x", name="labels")
 
     if chunks:
+        if not has_dask:
+            pytest.skip()
         array = array.chunk({"x": chunks})
         labels = labels.chunk({"x": chunks})
 
@@ -279,6 +274,7 @@ def test_xarray_groupby_bins(chunks):
             labels,
             dim="x",
             func="count",
+            engine=engine,
             expected_groups=np.array([1, 2, 4, 5]),
             isbin=True,
             fill_value=0,
@@ -291,3 +287,19 @@ def test_xarray_groupby_bins(chunks):
     xr.testing.assert_equal(actual, expected)
 
     # TODO: test cut_kwargs
+
+
+@requires_dask
+def test_func_is_aggregation():
+    from flox.aggregations import mean
+
+    ds = xr.tutorial.open_dataset("rasm", chunks={"time": 9})
+    expected = xarray_reduce(ds.Tair, ds.time.dt.month, func="mean")
+    actual = xarray_reduce(ds.Tair, ds.time.dt.month, func=mean)
+    xr.testing.assert_allclose(actual, expected)
+
+    with pytest.raises(ValueError):
+        xarray_reduce(ds.Tair, ds.time.dt.month, func=mean, skipna=True)
+
+    with pytest.raises(ValueError):
+        xarray_reduce(ds.Tair, ds.time.dt.month, func=mean, skipna=False)
